﻿Option Explicit On

Imports Excel = Microsoft.Office.Interop.Excel
Imports Microsoft.Office.Tools.Excel
Imports System.Data
Imports System.Data.OleDb
Imports System.Diagnostics

Module excel_sql_server

    Public LastErrorMessage As String

#Region "Create a QueryTable with Database data"

    ''' <summary>
    ''' The basic function to create ListObject with QueryTable.
    ''' </summary>
    ''' <param name="connectionString">The connection string for the QueryTable.</param>
    ''' <param name="commandText">The command text for the QueryTable.</param>
    ''' <param name="target">The target cell or range to insert the QueryTable.</param>
    ''' <returns>The status code.</returns>
    ''' <remarks>
    ''' The target range ListObject, if exists, is deleted before insert.
    ''' </remarks>
    Public Function CreateQueryTable(ByVal connectionString As String,
        ByVal commandText As String, ByVal target As Excel.Range) As Integer

        On Error Resume Next

        Dim ws As Excel.Worksheet = target.Worksheet

        If Not target.ListObject Is Nothing Then
            ws.ShowAllData()               ' Excel has a bug of the filtered ListObject deletion
            target.ListObject.Delete()
        End If

        ' Excel QueryTable requires "OLEDB;" at the start of the connection string

        connectionString = "OLEDB;" & connectionString

        ' Connection string of QueryTable has length limit

        Dim source(1) As Object
        source(0) = Mid(connectionString, 1, 200)
        source(1) = Mid(connectionString, 201)

        Dim lo As Excel.ListObject =
            ws.ListObjects.Add(SourceType:=0, Source:=source, Destination:=target.Cells(1, 1))

        ' lo.DisplayName = ws.Name & "_Table"

        With lo.QueryTable
            Select Case UCase(Left(Trim(commandText), 4))
                Case "EXEC", "SELE"
                    .CommandType = Excel.XlCmdType.xlCmdSql
                Case Else
                    .CommandType = Excel.XlCmdType.xlCmdTable
            End Select

            .CommandText = commandText
            .SavePassword = True
            '.RowNumbers = True
            .PreserveColumnInfo = True
            .PreserveFormatting = True
            .AdjustColumnWidth = False          ' Prevents column width change after update

            .Refresh(BackgroundQuery:=False)
        End With

        ' AutoFitRange(lo.Range)

        CreateQueryTable = 0
        LastErrorMessage = Nothing

    End Function

#End Region

#Region "Import Database Data to a Range using SQL Command"

    ''' <summary>
    ''' The basic function to copy database data to a range.
    ''' </summary>
    ''' <param name="connectionString">The connection string for OLEDB data source.</param>
    ''' <param name="commandText">The command text for OLEDB data source.</param>
    ''' <param name="target">The target cell or range to insert the data.</param>
    ''' <returns>The status code.</returns>
    ''' <remarks>
    ''' A ListObject is created for imported data.
    ''' </remarks>
    Public Function ImportSQLtoRange(ByVal connectionString As String,
        ByVal commandText As String, ByVal target As Excel.Range) As Integer

        Try
            Dim ws As Excel.Worksheet = target.Worksheet

            Dim arr(0, 0) As Object

            ExecuteOleDbDataReader(connectionString, commandText, AddressOf OleDbDataReaderToArray, arr)

            Dim rowCount As Long = arr.GetLength(1)
            Dim colCount As Integer = arr.GetLength(0)

            Dim r As Excel.Range            ' The complete target range with arr dimensions
            With ws
                r = .Application.Range(
                    .Cells(target.Row, target.Column),
                    .Cells(target.Row + rowCount - 1, target.Column + colCount - 1)
                )
            End With

            ' This method works very fast

            r.Value = ws.Application.WorksheetFunction.Transpose(arr)

            ws.ListObjects.Add(Excel.XlListObjectSourceType.xlSrcRange, r, , Excel.XlYesNoGuess.xlYes)

            ' AutoFitRange(r)

            ImportSQLtoRange = 0
            LastErrorMessage = Nothing

        Catch ex As Exception
            ImportSQLtoRange = 1
            LastErrorMessage = ex.Message

        End Try

    End Function

#End Region

#Region "Export an Excel Range to a Database"

    ''' <summary>
    ''' The basic function to export range data to a database.
    ''' </summary>
    ''' <param name="sourceRange">The source data range.</param>
    ''' <param name="connectionString">The OLEDB connection string for a target database.</param>
    ''' <param name="tableName">The target database table name.</param>
    ''' <param name="beforeSQL">The optional SQL code that executed before insert.</param>
    ''' <param name="afterSQL">The optional SQL code that executed after insert.</param>
    ''' <returns>The status code.</returns>
    ''' <remarks></remarks>
    <System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Security",
        "CA2100:Review SQL queries for security vulnerabilities")>
    Public Function ExportRangeToDatabase(ByVal sourceRange As Excel.Range,
        ByVal connectionString As String, ByVal tableName As String,
        Optional ByVal beforeSQL As String = "",
        Optional ByVal afterSQL As String = "") As Integer

        Dim commandText As String = "SELECT * FROM " & tableName
        Dim cmd As OleDbCommand

        Try
            Using connection As New OleDbConnection(connectionString)

                If beforeSQL > "" Then
                    cmd = New OleDbCommand(beforeSQL, connection)
                    Try
                        cmd.CommandType = CommandType.Text
                        connection.Open()
                        cmd.ExecuteNonQuery()
                    Finally
                        If Not cmd Is Nothing Then : cmd.Dispose() : End If
                    End Try
                End If

                Dim dataAdapter As New OleDbDataAdapter(commandText, connection)
                Try
                    Dim dataTable As New DataTable
                    Try
                        dataAdapter.Fill(dataTable)

                        Dim builder As OleDbCommandBuilder =
                            New OleDbCommandBuilder(dataAdapter)
                        Try
                            dataAdapter.InsertCommand = builder.GetInsertCommand()

                            ' Column mappings

                            Dim rangeFields(100) As Integer
                            Dim tableFields(100) As Integer
                            Dim tableFieldTypes(100) As String

                            Dim exportFieldsCount As Integer
                            exportFieldsCount = 0

                            Dim col As Integer
                            Dim index As Integer

                            ' Set the range before .Match call!

                            Dim rangeHeaders As Excel.Range = sourceRange.Rows(1)

                            Dim cellFormat As String

                            For col = 0 To dataTable.Columns.Count - 1
                                index = sourceRange.Application.Match(
                                    dataTable.Columns(col).ColumnName, rangeHeaders, 0)
                                If index > 0 Then
                                    exportFieldsCount = exportFieldsCount + 1
                                    rangeFields(exportFieldsCount) = index
                                    tableFields(exportFieldsCount) = col
                                    tableFieldTypes(exportFieldsCount) =
                                        dataTable.Columns(col).DataType.ToString

                                    ' Transformation for the Time datatype

                                    If tableFieldTypes(exportFieldsCount) = "System.String" Then
                                        cellFormat =
                                            sourceRange.Cells(2, index).NumberFormat
                                        If cellFormat.Contains(":") Then
                                            tableFieldTypes(exportFieldsCount) = "Time"
                                        End If
                                    End If
                                End If
                            Next

                            If exportFieldsCount = 0 Then
                                ExportRangeToDatabase = 1
                                Exit Function
                            End If

                            ' Fast read of Excel range values to an array
                            ' for further fast work with the array

                            Dim arr As Object
                            arr = sourceRange.Value

                            ' Column names should be equal
                            ' For col = 1 To exportFieldsCount
                            '     Debug.Print(dataTable.Columns(tableFields(col)).ColumnName & " = " & arr(1, rangeFields(col)))
                            ' Next

                            ' The range data transfer to the dataTable

                            Dim row As Long
                            Dim rowCount As Long
                            rowCount = UBound(arr, 1)

                            Dim value As Object

                            Dim dataRow As DataRow

                            For row = 2 To rowCount
                                dataRow = dataTable.NewRow()
                                For col = 1 To exportFieldsCount
                                    value = arr(row, rangeFields(col))
                                    If value Is Nothing Then
                                    ElseIf tableFieldTypes(col) = "System.DateTime" Then
                                        If value.GetType.ToString() = "System.Double" Then
                                            dataRow(tableFields(col)) = DateTime.FromOADate(value)
                                        Else
                                            dataRow(tableFields(col)) = value
                                        End If
                                    ElseIf tableFieldTypes(col) = "Time" Then
                                        If value.GetType.ToString() = "System.Double" Then
                                            dataRow(tableFields(col)) = DateTime.FromOADate(value).ToString("HH:mm:ss")
                                        Else
                                            dataRow(tableFields(col)) = DateTime.Parse(value).ToString("HH:mm:ss")
                                        End If
                                    Else
                                        dataRow(tableFields(col)) = value
                                    End If
                                Next
                                dataTable.Rows.Add(dataRow)
                            Next

                            dataAdapter.Update(dataTable)

                        Finally
                            builder.Dispose()
                        End Try
                    Finally
                        dataTable.Dispose()
                    End Try
                Finally
                    dataAdapter.Dispose()
                End Try

                If afterSQL > "" Then
                    cmd = New OleDbCommand(afterSQL, connection)
                    Try
                        cmd.CommandType = CommandType.Text
                        cmd.ExecuteNonQuery()
                    Finally
                        If Not cmd Is Nothing Then : cmd.Dispose() : End If
                    End Try
                End If
            End Using

            ExportRangeToDatabase = 0
            LastErrorMessage = Nothing

        Catch sqlExc As OleDbException
            ExportRangeToDatabase = 1
            LastErrorMessage = sqlExc.Message

        Catch ex As Exception
            ExportRangeToDatabase = 1
            LastErrorMessage = ex.Message

        End Try

    End Function

#End Region

#Region "Tools ListObject Import-Export"

    ' The global DataAdapters list
    Public DataAdapters As New SortedList(Of String, OleDbDataAdapter)

    ''' <summary>
    ''' The basic function to create Tools ListObject connected to a database
    ''' </summary>
    ''' <param name="connectionString">The connection string for OLEDB data source.</param>
    ''' <param name="commandText">The command text for OLEDB data source.</param>
    ''' <param name="sh">The Tools Worksheet.</param>
    ''' <param name="target">The target Excel range to insert the Tools ListObject.</param>
    ''' <returns>The status code.</returns>
    ''' <remarks></remarks>
    <System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Security",
        "CA2100:Review SQL queries for security vulnerabilities")>
    Function ImportToolsListObject(ByVal connectionString As String, ByVal commandText As String,
        ByVal sh As WorksheetBase, ByVal target As Excel.Range) As Integer

        Dim listObjectName As String = String.Format("DataTable_{0}", DataAdapters.Count + 1)

        Try
            Using connection As New OleDbConnection(connectionString)

                Dim dataTable As New DataTable
                Dim dataAdapter As New OleDbDataAdapter(commandText, connection)

                dataAdapter.Fill(dataTable)

                Dim builder As OleDbCommandBuilder = New OleDbCommandBuilder(dataAdapter)
                With dataAdapter
                    .InsertCommand = builder.GetInsertCommand()
                    .UpdateCommand = builder.GetUpdateCommand()
                    .DeleteCommand = builder.GetDeleteCommand()
                End With

                DataAdapters.Add(listObjectName, dataAdapter)

                With sh.Controls.AddListObject(target, listObjectName)
                    .AutoSetDataBoundColumnHeaders = True
                    .DataSource = dataTable
                End With

                ' Don't call Dispose for these objects:
                ' builder.Dispose()
                ' dataAdapter.Dispose()
                ' dataTable.Dispose()
            End Using

            ImportToolsListObject = 0
            LastErrorMessage = Nothing

        Catch ex As Exception
            ImportToolsListObject = 1
            LastErrorMessage = ex.Message

        End Try

    End Function

    ''' <summary>
    ''' The basic function to export Tools ListObject data changes to a database.
    ''' </summary>
    ''' <param name="connectionString">The connection string for OLEDB data source.</param>
    ''' <param name="lo">The Tools ListObject.</param>
    ''' <returns>The status code.</returns>
    ''' <remarks>
    ''' The Tools ListObject data are updated fully after the data export.
    ''' DataAdapters should contain a DataAdapter for the Tools ListObject.
    ''' </remarks>
    Function ExportToolsListObject(ByVal connectionString As String,
        ByVal lo As Microsoft.Office.Tools.Excel.ListObject) As Integer

        If DataAdapters.Count = 0 Then
            ExportToolsListObject = 1
            LastErrorMessage = "There are no DataAdapters"
            Exit Function
        End If

        Try
            Dim dataTable As DataTable = lo.DataSource

            Using connection As New OleDbConnection(connectionString)
                With DataAdapters(lo.DisplayName)

                    .SelectCommand.Connection = connection
                    .InsertCommand.Connection = connection
                    .UpdateCommand.Connection = connection
                    .DeleteCommand.Connection = connection

                    .Update(dataTable)

                    dataTable.Clear()

                    .Fill(dataTable)
                End With
            End Using

            ExportToolsListObject = 0
            LastErrorMessage = Nothing

        Catch ex As Exception
            ExportToolsListObject = 1
            LastErrorMessage = ex.Message

        End Try

    End Function

    ''' <summary>
    ''' The basic function to refresh Tools ListObject data from a database.
    ''' </summary>
    ''' <param name="connectionString">The connection string for OLEDB data source.</param>
    ''' <param name="lo">The Tools ListObject.</param>
    ''' <returns>The status code.</returns>
    ''' <remarks>
    ''' DataAdapters should contain a DataAdapter for the Tools ListObject.
    ''' </remarks>
    Function RefreshToolsListObject(ByVal connectionString As String,
        ByVal lo As Microsoft.Office.Tools.Excel.ListObject) As Integer

        If DataAdapters.Count = 0 Then
            RefreshToolsListObject = 1
            LastErrorMessage = "There are no DataAdapters"
            Exit Function
        End If

        Try
            Dim dataTable As DataTable = lo.DataSource

            dataTable.Clear()

            Using connection As New OleDbConnection(TestConnectionString)
                With DataAdapters(lo.DisplayName)

                    .SelectCommand.Connection = connection
                    .Fill(dataTable)
                End With
            End Using

            RefreshToolsListObject = 0
            LastErrorMessage = Nothing

        Catch ex As Exception
            RefreshToolsListObject = 1
            LastErrorMessage = ex.Message

        End Try

    End Function

#End Region

#Region "Useful Excel Procedures"

    ''' <summary>
    ''' The procedure to auto-fit the column widths of an input range.
    ''' </summary>
    ''' <param name="r">The range for column widths auto-fit.</param>
    ''' <remarks></remarks>
    Public Sub AutoFitRange(ByVal r As Excel.Range)

        For Each column As Excel.Range In r.Columns
            column.EntireColumn.AutoFit()
            If column.ColumnWidth > 150 Then
                column.ColumnWidth = 150
            End If
        Next

    End Sub

#End Region

#Region "Connection String Functions"

    ''' <summary>
    ''' The function makes a connection string for OLEDB data source.
    ''' </summary>
    ''' <param name="server">The server.</param>
    ''' <param name="database">The database.</param>
    ''' <param name="username">The username. It should be empty for the Windows-authentication.</param>
    ''' <param name="password">The password. It is ignored for empty username.</param>
    ''' <param name="extended">The extended properties string.</param>
    ''' <returns>The OLEDB connection string.</returns>
    ''' <remarks>
    ''' If the server contains ".NET" (SQL Azure) and extended is ""
    ''' then extended properties filled with SQL Azure required properties.
    ''' </remarks>
    Public Function GetConnectionString(ByVal server As String, ByVal database As String,
                ByVal username As String, ByVal password As String,
                Optional ByVal extended As String = "") As String

        If (extended = "") And UCase(server).Contains(".NET") Then
            extended = "Extended Properties=""Encrypt=True;TrustServerCertificate=False"";"
        End If

        If username = "" Then
            Return String.Format("Provider=SQLOLEDB.1;Integrated Security=SSPI{2};" &
                                 "Persist Security Info=True;" &
                                 "Data Source={0};Use Procedure for Prepare=1;Auto Translate=True;" &
                                 "Packet Size=4096;Initial Catalog={1}",
                                    server, database, extended)
        Else
            Return String.Format("Provider=SQLOLEDB.1;Password={3};User ID={2};{4}" &
                                 "Persist Security Info=True;" &
                                 "Data Source={0};Use Procedure for Prepare=1;Auto Translate=True;" &
                                 "Packet Size=4096;Initial Catalog={1}",
                                    server, database, username, password, extended)
        End If

    End Function

#End Region

#Region "OleDbDataReader Procedures"

    ''' <summary>
    ''' The OleDbDataReader processor delegate.
    ''' </summary>
    ''' <param name="dataReader">The opened OleDbDataReader.</param>
    ''' <param name="context">The input/output context object.</param>
    ''' <remarks>
    ''' The delegate is used in ExecuteOleDbDataReader function to process
    ''' OleDbDataReader with input/output context object.
    ''' </remarks>
    Delegate Sub ExecuteOleDbDataReaderDelegate(ByVal dataReader As OleDbDataReader, ByRef context As Object)

    ''' <summary>
    ''' The universal function to process OleDbDataReader using processor delegate and context object.
    ''' </summary>
    ''' <param name="connectionString">The OLEDB connection string.</param>
    ''' <param name="commandText">The command text for the ExecuteReader OleDbCommand.</param>
    ''' <param name="processor">The OleDbDataReader processor.</param>
    ''' <param name="context">The input/output context object.</param>
    ''' <remarks>
    ''' The function opens a connection, executes an ExecuteReader command,
    ''' calls the processor delegate with opened OleDbDataReader and the context object
    ''' and closes the connection.
    ''' </remarks>
    <System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Security",
        "CA2100:Review SQL queries for security vulnerabilities")>
    Public Function ExecuteOleDbDataReader(ByVal connectionString As String, ByVal commandText As String,
        ByVal processor As ExecuteOleDbDataReaderDelegate, ByRef context As Object) As Boolean

        Try
            Using connection As New OleDbConnection(connectionString)

                Dim cmd As New OleDbCommand(commandText, connection)
                Try
                    cmd.CommandType = CommandType.Text
                    connection.Open()
                    Using dataReader As OleDbDataReader = cmd.ExecuteReader(CommandBehavior.CloseConnection)

                        processor(dataReader, context)

                    End Using
                Finally
                    If Not cmd Is Nothing Then : cmd.Dispose() : End If
                End Try
                Return True
            End Using
        Catch sqlExc As OleDbException
            MsgBox(sqlExc.Message, MsgBoxStyle.Exclamation)
            Return False
        Catch ex As Exception
            MsgBox(ex.Message, MsgBoxStyle.Exclamation)
            Return False
        End Try

    End Function

    ''' <summary>
    ''' The function to load an OleDbDataReader to a DataTable context.
    ''' </summary>
    ''' <param name="dataReader">The opened OleDbDataReader.</param>
    ''' <param name="context">The DataTable for OleDbDataReader load.</param>
    ''' <remarks>
    ''' The function is used as a delegate for ExecuteOleDBDataReader function.
    ''' </remarks>
    Public Sub OleDbDataReaderToDataTable(ByVal dataReader As OleDbDataReader, ByRef context As Object)

        Dim dt As DataTable = context
        dt.Load(dataReader, LoadOption.PreserveChanges)

    End Sub

    ''' <summary>
    ''' The function to load an OleDbDataReader to an array context.
    ''' </summary>
    ''' <param name="dataReader">The opened OleDbDataReader.</param>
    ''' <param name="context">The output array of headers and data reader data.</param>
    ''' <remarks>
    ''' The function is used as a delegate for ExecuteOleDBDataReader function.
    ''' The context should be declared as Dim arr(0, 0) As Object
    ''' The output array contains transposed data of the source data reader.
    ''' The first row of the array contains column header.
    ''' </remarks>
    Public Sub OleDbDataReaderToArray(ByVal dataReader As OleDbDataReader, ByRef context As Object)

        Dim row As Long = -1
        Dim colCount As Integer = dataReader.FieldCount
        Dim rowCapacity As Long = 1000

        ReDim context(colCount - 1, rowCapacity - 1)

        row = row + 1
        For col As Integer = 0 To colCount - 1
            context(col, row) = dataReader.GetName(col)
        Next

        Do While dataReader.Read
            row = row + 1
            If row = rowCapacity Then
                rowCapacity = rowCapacity + 1000
                ReDim Preserve context(colCount - 1, rowCapacity - 1)
            End If
            For col As Integer = 0 To colCount - 1
                context(col, row) = dataReader(col).ToString
            Next
        Loop

        ReDim Preserve context(colCount - 1, row)

    End Sub

#End Region

End Module
